<?php

namespace Drupal\url_status_scanner\Form;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\StatementInterface;
use Drupal\Core\Extension\ExtensionList;
use Drupal\Core\Form\FormStateInterface;
use Drupal\url_status_scanner\Service\UrlStatusScannerService;
use PhpOffice\PhpSpreadsheet\Spreadsheet;
use PhpOffice\PhpSpreadsheet\Writer\Xlsx;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * This form is used to display the dashboard for the URL Status Scanner module.
 */
class UrlStatusScannerDashboard {

  /**
   * Get the  urlStatusScannerService.
   *
   * @var \Drupal\url_status_scanner\Service\UrlStatusScannerService
   */
  protected $urlStatusScannerService;

  /**
   * Get the database connection object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The module extension list service.
   *
   * @var \Drupal\Core\Extension\ExtensionList
   */
  protected $extensionList;

  /**
   * Constructs a new UrlStatusScannerDashboard object.
   *
   * @param \Drupal\url_status_scanner\Service\UrlStatusScannerService $url_status_scanner_service
   *   The urlStatusScannerService.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Extension\ExtensionList $extension_list
   *   The module extension list service.
   */
  public function __construct(UrlStatusScannerService $url_status_scanner_service, Connection $database, ExtensionList $extension_list) {
    $this->urlStatusScannerService = $url_status_scanner_service;
    $this->database = $database;
    $this->extensionList = $extension_list;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('url_status_scanner.service'),
      $container->get('database'),
      $container->get('extension.list.module')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'url_status_scanner_dashboard';
  }

  /**
   * {@inheritdoc}
   *
   * @param array<string, mixed> $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array<string, mixed>
   *   The form structure.
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {

    // Show an intro paragraph.
    $form['intro'] = [
      '#type' => 'html_tag',
      '#tag' => 'p',
      '#value' => $this->t('The URL Status Scanner module helps track the status of external URLs.'),
    ];

    // Add the filters.
    $form['filters'] = [
      '#type' => 'details',
      '#title' => $this->t('Filter results'),
      '#open' => TRUE,
    ];

    $form['filters']['filter_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('URL'),
      '#default_value' => $form_state->getValue('filter_url', ''),
      '#description' => $this->t('Enter a URL to filter results by. Leave blank to show all results.'),
    ];

    $form['filters']['filter_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Type'),
      '#options' => [
        '' => $this->t('- Any -'),
        'Basecode' => $this->t('Base code'),
        'In content' => $this->t('In content'),
        'In config' => $this->t('In config'),
      ],
      '#default_value' => $form_state->getValue('filter_type', ''),
      '#description' => $this->t('Select a type to filter results by. <ul><li><strong>Base code:</strong> The URLs are in the base code files from custom modules and themes.</li><li><strong>In content:</strong> The URLs are in the content entities, such as nodes, comments, etc.</li><li><strong>In config:</strong> The URLs are in the config objects.</li></ul>'),
    ];

    $form['filters']['filter_status_code'] = [
      '#type' => 'select',
      '#title' => $this->t('Status Code'),
      '#options' => [
        '' => $this->t('- Any -'),
        '200' => $this->t('200 OK'),
        '301' => $this->t('301 Moved Permanently'),
        '302' => $this->t('302 Moved Temporarily'),
        '401' => $this->t('401 Unauthorized'),
        '403' => $this->t('403 Forbidden'),
        '404' => $this->t('404 Not Found'),
        '408' => $this->t('408 Request Timeout'),
        '500' => $this->t('500 Internal Server Error'),
        '503' => $this->t('503 Service Unavailable'),
      ],
      '#default_value' => $form_state->getValue('filter_status_code', ''),
      '#description' => $this->t('Select a status code to filter results by:<br><ul><li>200 OK</li><li>301 Moved Permanently</li><li>302 Moved Temporarily</li><li>401 Unauthorized</li><li>403 Forbidden</li><li>404 Not Found</li><li>408 Request Timeout</li><li>500 Internal Server Error</li><li>503 Service Unavailable</li></ul>'),
    ];

    $form['filters']['actions'] = [
      '#type' => 'actions',
    ];

    $form['filters']['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Apply filters'),
      '#submit' => ['::applyFilters'],
    ];

    $form['filters']['actions']['reset'] = [
      '#type' => 'submit',
      '#value' => $this->t('Reset'),
      '#submit' => ['::resetFilters'],
      '#limit_validation_errors' => [],
    ];

    // Set the table header.
    $header = [
      'url' => [
        'data' => $this->t('URL'),
        'field' => 'url',
        'specifier' => 'url',
      ],
      'location' => [
        'data' => $this->t('Location'),
        'field' => 'location',
        'specifier' => 'location',
      ],
      'type' => [
        'data' => $this->t('Type'),
        'field' => 'type',
        'specifier' => 'type',
      ],
      'status' => [
        'data' => $this->t('Status'),
        'field' => 'status_code',
        'specifier' => 'status_code',
      ],
    ];

    // Get the values from the database.
    $query = $this->database->select('url_status_scanner', 'u')
      ->fields('u', ['url', 'location', 'type', 'status_code']);

    // Apply sorting.
    $query = $query->extend('\Drupal\Core\Database\Query\TableSortExtender')
      ->orderByHeader($header);

    // Apply limit results by page.
    $query = $query->extend('\Drupal\Core\Database\Query\PagerSelectExtender')
      ->limit(25);

    // Apply filters.
    if ($form_state->getValue('filter_url')) {
      $query->condition('u.url', '%' . $form_state->getValue('filter_url') . '%', 'LIKE');
    }

    if ($form_state->getValue('filter_type')) {
      $query->condition('u.type', $form_state->getValue('filter_type'));
    }

    if ($form_state->getValue('filter_status_code')) {
      $query->condition('u.status_code', $form_state->getValue('filter_status_code'));
    }

    $results = $query->execute()->fetchAll();

    // Build the table.
    $form['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => array_map(fn($r) => [
        'data' => [
          ['data' => (string) $r->url],
          ['data' => (string) $r->location],
          ['data' => (string) $r->type],
          ['data' => (string) $r->status_code],
        ],
      ], $results),
      '#empty' => $this->t('No results found.'),
    ];

    // Add the library to the form.
    if (!isset($form['#attached']) || !is_array($form['#attached'])) {
      $form['#attached'] = [];
    }
    if (!isset($form['#attached']['library']) || !is_array($form['#attached']['library'])) {
      $form['#attached']['library'] = [];
    }

    $form['#attached']['library'][] = 'url_status_scanner/form_status';

    $form['process_message'] = [
      '#markup' => '<div id="process-message" style="display:none; color: red; font-weight: bold;">' .
      $this->t('Processing, please wait...') . '</div>',
    ];

    // Add the submit button.
    $form['button'] = [
      '#type' => 'submit',
      '#value' => $this->t('Scan URLs'),
      '#attributes' => ['id' => 'submit-button'],
    ];

    // Add the pager.
    $form['pager'] = [
      '#type' => 'pager',
    ];

    // Add the data export section.
    $form['export'] = [
      '#type' => 'details',
      '#title' => $this->t('Export data'),
      '#open' => TRUE,
    ];

    $form['export']['export_csv'] = [
      '#type' => 'submit',
      '#value' => $this->t('Export to CSV'),
      '#submit' => ['::exportCsv'],
    ];

    $form['export']['export_xlsx'] = [
      '#type' => 'submit',
      '#value' => $this->t('Export to XLSX'),
      '#submit' => ['::exportXlsx'],
    ];

    return $form;
  }

  /**
   * Submit handler for the filter form.
   *
   * @param array<string, mixed> $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function applyFilters(array &$form, FormStateInterface $form_state): void {

    $values = $form_state->getValues();

    // Trim the values to remove leading and trailing spaces.
    if (!empty($values['filter_url'])) {
      $form_state->setValue('filter_url', trim($values['filter_url']));
    }

    $form_state->setRebuild(TRUE);
  }

  /**
   * Reset handler for the filter form.
   *
   * @param array<string, mixed> $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function resetFilters(array &$form, FormStateInterface $form_state): void {
    $form_state->setValues([]);
    $form_state->setRedirect('<current>');
  }

  /**
   * {@inheritdoc}
   *
   * @param array<string, mixed> $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {

    // Get all the URLs to scan.
    $urls = iterator_to_array($this->urlStatusScannerService->detectAllUrls(), FALSE);

    if (empty($urls)) {
      $this->messenger()->addWarning($this->t('No URLs found to scan.'));
      return;
    }

    $operations = [];
    foreach ($urls as $result) {
      $operations[] = [
        '\Drupal\url_status_scanner\Form\UrlStatusScannerDashboard::processBatch', [$result],
      ];
    }

    // Set up the batch process.
    $batch = [
      'title' => $this->t('Processing URLs...'),
      'operations' => $operations,
      'init_message' => $this->t('Starting URL processing...'),
      'progress_message' => $this->t('Processed @current out of @total.'),
      'error_message' => $this->t('An error occurred while processing URLs.'),
      'finished' => 'finishBatch',
      'file' => $this->extensionList->getPath('url_status_scanner') . '/src/Form/UrlStatusScannerDashboard.php',
    ];

    // Run the batch process.
    batch_set($batch);

    // Redirect to the dashboard adding the query parameters for sorting.
    $query = [
      'order_id' => $form_state->getValue('order_id'),
      'sort' => $this->getRequest()->query->get('sort'),
      'order' => $this->getRequest()->query->get('order'),
    ];
    $form_state->setRedirect('url_status_scanner.dashboard', [], ['query' => $query]);
  }

  /**
   * Processes the batch item.
   *
   * @param array<string, mixed> $url
   *   Keyed array of data to process.
   * @param array<string, mixed> $context
   *   The batch context.
   */
  public static function processBatch($url, &$context): void {

    // Process the URL.
    $service = \Drupal::service('url_status_scanner.service');
    $service->checkUrl($url);

    // Store the results in the database.
    $service->saveUrlResults($url);

    // Show progress message.
    $context['message'] = t('Processing @url', ['@url' => $url['url']]);
  }

  /**
   * Finish batch.
   *
   * @param bool $success
   *   Indicates whether the batch process was successful.
   * @param array<int, mixed> $results
   *   Results information passed from the processing callback.
   * @param array<int, array<int, mixed>> $operations
   *   The batch operations.
   */
  public static function finishBatch($success, $results, $operations): void {
    if ($success) {
      \Drupal::messenger()->addStatus(t('All URLs processed successfully.'));
    }
    else {
      \Drupal::messenger()->addError(t('There was an issue processing some URLs.'));
    }
    \Drupal::service('cache.render')->invalidateAll();
  }

  /**
   * Export data to CSV.
   *
   * @param array<string, mixed> $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return void
   *   Returns nothing.
   */
  public function exportCsv(array &$form, FormStateInterface $form_state): void {
    $filename = 'url_status_report.csv';

    // Set the headers for downloading the CSV file.
    header('Content-Type: text/csv; charset=UTF-8');
    header('Content-Disposition: attachment; filename="' . $filename . '"');
    header('Pragma: no-cache');
    header('Expires: 0');

    // Open the output stream for writing.
    $output = fopen('php://output', 'w');

    if ($output === FALSE) {
      throw new \RuntimeException('Cannot open output stream for writing.');
    }

    // Overwrite the BOM for avoiding encoding issues in Excel.
    fwrite($output, "\xEF\xBB\xBF");

    // Add the CSV headers.
    fputcsv($output, ['URL', 'Location', 'Type', 'Status Code']);

    // Get the values from the database.
    $query = $this->database->select('url_status_scanner', 'u')
      ->fields('u', ['url', 'location', 'type', 'status_code']);
    $result = $query->execute();

    if (!$result instanceof StatementInterface) {
      throw new \RuntimeException('Error al ejecutar la consulta SQL.');
    }

    $rows = $result->fetchAll();

    // Write the data to the CSV file, encoding UTF-8.
    foreach ($rows as $row) {
      fputcsv($output, [
        mb_convert_encoding((string) $row->url, 'UTF-8', 'auto'),
        mb_convert_encoding((string) $row->location, 'UTF-8', 'auto'),
        mb_convert_encoding((string) $row->type, 'UTF-8', 'auto'),
        mb_convert_encoding((string) $row->status_code, 'UTF-8', 'auto'),
      ]);
    }

    fclose($output);
    exit();
  }

  /**
   * Export data to XLSX.
   *
   * @param array<string, mixed> $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return void
   *   Returns nothing.
   */
  public function exportXlsx(array &$form, FormStateInterface $form_state): void {
    $filename = 'url_status_report.xlsx';

    // Get the values from the database.
    $query = $this->database->select('url_status_scanner', 'u')
      ->fields('u', ['url', 'location', 'type', 'status_code']);
    $results = $query->execute();

    if (!$results instanceof StatementInterface) {
      throw new \RuntimeException('Error al ejecutar la consulta SQL.');
    }

    // Fetch the results.
    $data = $results->fetchAll();

    // Create a new spreadsheet.
    $spreadsheet = new Spreadsheet();
    $sheet = $spreadsheet->getActiveSheet();

    // Add headers.
    $headers = ['URL', 'Location', 'Type', 'Status Code'];
    $sheet->fromArray([$headers], NULL, 'A1');

    // Add data.
    $row = 2;
    foreach ($data as $result) {
      $sheet->fromArray([
        (string) $result->url,
        (string) $result->location,
        (string) $result->type,
        (string) $result->status_code,
      ], NULL, "A$row");
      $row++;
    }

    // Generate the XLSX file, and send it to the browser.
    header('Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet');
    header('Content-Disposition: attachment; filename="' . $filename . '"');
    header('Cache-Control: max-age=0');

    $writer = new Xlsx($spreadsheet);
    $writer->save('php://output');
    exit();
  }

}
