<?php

namespace Drupal\url_status_scanner\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\RequestException;

/**
 * This service detects URLs in all sources and checks their status.
 */
class UrlStatusScannerService {

  /**
   * Get the database connection object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Get the config factory object.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * Constructs a new UrlStatusScannerService object.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity
   *   The entity object.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle service.
   */
  public function __construct(
    Connection $database,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
  ) {
    $this->database = $database;
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * Detect URL's in all sources.
   */
  public function detectAllUrls(): \Generator {

    if ($this->database->schema()->tableExists('url_status_scanner')) {
      $this->database->truncate('url_status_scanner')->execute();
    }

    // Gather all the URLs from all methods using `yield`.
    yield from $this->detectUrlsInBasecode();
    yield from $this->detectUrlsInConfig();
    yield from $this->detectUrlsInContent();
  }

  /**
   * Method 1: Detect URLs in custom code files.
   */
  public function detectUrlsInBasecode(): \Generator {
    $config_scan = $this->configFactory->get('url_status_scanner.settings')->get('scan_basecode');

    // If the option is disabled in the settings form,
    // finish the method execution.
    if (!$config_scan || $config_scan == FALSE) {
      return;
    }

    // Paths to scan in a Drupal Standard installation.
    $paths_to_scan = [
      DRUPAL_ROOT . '/modules/custom',
      DRUPAL_ROOT . '/themes/custom',
    ];

    // Add `sites/default` if it has custom modules or themes.
    if (is_dir(DRUPAL_ROOT . '/sites/default/modules/custom')) {
      $paths_to_scan[] = DRUPAL_ROOT . '/sites/default/modules/custom';
    }
    if (is_dir(DRUPAL_ROOT . '/sites/default/themes/custom')) {
      $paths_to_scan[] = DRUPAL_ROOT . '/sites/default/themes/custom';
    }

    // Detect and add custom modules and themes in a Drupal Multisite.
    $site_directories = glob(DRUPAL_ROOT . '/sites/*', GLOB_ONLYDIR) ?: [];
    foreach ($site_directories as $site_directory) {
      // Avoid to duplicate `sites/default`.
      if ($site_directory !== DRUPAL_ROOT . '/sites/default') {
        if (is_dir($site_directory . '/modules/custom')) {
          $paths_to_scan[] = $site_directory . '/modules/custom';
        }
        if (is_dir($site_directory . '/themes/custom')) {
          $paths_to_scan[] = $site_directory . '/themes/custom';
        }
      }
    }

    // Iterate over all directories to scan.
    foreach ($paths_to_scan as $directory) {
      if (!is_dir($directory)) {
        continue;
      }

      // Iterate over all files in the directory.
      try {
        $iterator = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($directory));
      }
      catch (\UnexpectedValueException $e) {
        continue;
      }

      foreach ($iterator as $fileInfo) {

        if (!$fileInfo instanceof \SplFileInfo) {
          continue;
        }
        // Set the file extension to check.
        if (in_array($fileInfo->getExtension(), ['php', 'yml', 'twig', 'theme', 'module', 'css', 'js'])) {
          $lineNumber = 0;
          $handle = fopen($fileInfo->getRealPath(), 'r');
          // Iterate over all lines in the file.
          if ($handle !== FALSE) {
            while (($line = fgets($handle)) !== FALSE) {
              $lineNumber++;
              preg_match_all('/https?:\/\/[a-zA-Z0-9\-\.]+\.[a-zA-Z]{2,6}(\/[^\s\'")]*)?/', $line, $matches);
              if (!empty($matches[0])) {
                foreach ($matches[0] as $url) {
                  yield [
                    'url' => $url,
                    'location' => $fileInfo->getRealPath() . ':' . $lineNumber,
                    'type' => 'Basecode',
                  ];
                }
              }
            }
            fclose($handle);
          }
        }
      }
    }
  }

  /**
   * Method 2: Detect URLs in configuration settings.
   */
  public function detectUrlsInConfig(): \Generator {

    $config_name = "";
    $config_scan = $this->configFactory->get('url_status_scanner.settings')->get('scan_in_config');

    // If the option is disabled in the settings form,
    // finish the method execution.
    if (!$config_scan || $config_scan == FALSE) {
      return;
    }

    // Get all configuration settings.
    $urls = [];
    $all_configs = $this->configFactory->listAll();

    // Iterate over all configuration settings.
    foreach ($all_configs as $config_name) {
      $config = $this->configFactory->get($config_name)->getRawData();
      array_walk_recursive($config, function ($value) use (&$urls) {
        if (is_string($value) && preg_match('/^https?:\/\//', $value)) {
          $urls[] = $value;
        }
      });

      $urls = array_unique($urls);
      foreach ($urls as $url) {
        yield [
          'url' => $url,
          'location' => 'Configuration name: ' . $config_name,
          'type' => 'In config',
        ];
      }
    }
  }

  /**
   * Method 3: Detect URLs in content fields.
   */
  public function detectUrlsInContent(): \Generator {
    $config_scan = $this->configFactory->get('url_status_scanner.settings')->get('scan_in_content');

    // If the option is disabled in the settings form,
    // finish the method execution.
    if (!$config_scan || $config_scan == FALSE) {
      return;
    }

    // Get all entity types definitions.
    $entity_type_definitions = $this->entityTypeManager->getDefinitions();
    $batch_size = 5;

    // Iterate over all entity types.
    foreach ($entity_type_definitions as $entity_type_id => $definition) {
      if (!$definition instanceof ContentEntityType) {
        continue;
      }

      if ($entity_type_id === 'node') {
        $bundles = $this->entityTypeBundleInfo->getBundleInfo('node');
        foreach (array_keys($bundles) as $bundle) {
          yield from $this->scanBundle($entity_type_id, $bundle, $batch_size);
        }
      }
      else {
        yield from $this->scanBundle($entity_type_id, NULL, $batch_size);
      }
    }
  }

  /**
   * Summary of scanBundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The bundle.
   * @param int $batch_size
   *   The batch size.
   *
   * @return \Generator
   *   The scan bundle.
   */
  private function scanBundle(string $entity_type_id, ?string $bundle, int $batch_size): \Generator {
    $storage = $this->entityTypeManager->getStorage($entity_type_id);
    $query = $storage->getQuery()
      ->accessCheck(FALSE)
    // Limit the number of results to 200.
      ->range(0, 200);

    if ($bundle) {
      // Filter by bundle if provided.
      $query->condition('type', $bundle);
    }

    $entity_ids = $query->execute();
    if (empty($entity_ids)) {
      // Return an empty generator if no entities are found.
      return (function () {
        yield from [];
      })();
    }

    // Iterate over all entities and extract URLs.
    foreach (array_chunk($entity_ids, max(1, min($batch_size, 50))) as $chunk) {
      $entities = $storage->loadMultiple($chunk);

      foreach ($entities as $entity) {
        foreach ($entity->getFields() as $field_name => $field) {
          if ($field->isEmpty()) {
            continue;
          }

          $values = $field->getValue();
          foreach ($values as $value) {
            foreach ($value as $val) {
              if (is_string($val) && !empty($val)) {
                $urls = $this->extractUrlsFromHtml($val);
                foreach ($urls as $url) {
                  yield [
                    'url' => $url,
                    'location' => "Entity: $entity_type_id, Bundle: $bundle, ID: {$entity->id()}, Field: $field_name",
                    'type' => 'In content',
                  ];
                }
              }
            }
          }
        }
      }
      // Free memory.
      unset($entities);
      gc_collect_cycles();
    }
  }

  /**
   * Check a single URL and save the result.
   *
   * @param array<string, mixed> $url
   *   An associative array containing URL data.
   *
   * @return void
   *   This method does not return anything.
   */
  public function checkUrl(array &$url): void {
    $client = new Client([
      'timeout' => 10,
      'connect_timeout' => 5,
    ]);

    // Check if the URL is valid.
    if (!isset($url['url']) || !is_string($url['url']) || !filter_var($url['url'], FILTER_VALIDATE_URL)) {
      $url['status_code'] = 400;
      $this->saveUrlResults($url);
      return;
    }

    try {
      // Make the http request and get the status code.
      $response = $client->get($url['url']);
      $status_code = $response->getStatusCode();
    }
    catch (ConnectException $e) {
      // Get the CURL error message.
      $errorMessage = $e->getMessage();

      if (strpos($errorMessage, 'cURL error 7') !== FALSE) {
        $status_code = 503;
      }
      elseif (strpos($errorMessage, 'cURL error 28') !== FALSE) {
        $status_code = 408;
      }
      else {
        $status_code = 500;
      }
    }
    catch (RequestException $e) {
      $response = $e->getResponse();
      $status_code = $response ? $response->getStatusCode() : 500;
    }

    $url['status_code'] = $status_code;

    // Store the results in the database.
    $this->saveUrlResults($url);
  }

  /**
   * Save URL results in the database.
   *
   * @param array<string, mixed> $url
   *   An associative array containing URL data.
   *
   * @return void
   *   This method does not return anything.
   */
  public function saveUrlResults($url): void {
    $url_value = $url['url'] ?? '';
    if (!is_string($url_value)) {
      return;
    }

    // Truncate the URL if it is too long.
    $length = strlen($url_value);
    if ($length > 250) {
      $url['url'] = substr($url_value, 0, 250) . '...';
    }
    $urlProcess = strval($url_value);
    // Save the results in the database.
    $this->database->merge('url_status_scanner')
      ->key('url', $urlProcess)
      ->fields([
        'url' => $url['url'],
        'location' => $url['location'],
        'type' => $url['type'],
        'status_code' => $url['status_code'],
      ])
      ->execute();
  }

  /**
   * Extract URLs from HTML.
   *
   * @param string $html
   *   The HTML content from which to extract URLs.
   *
   * @return string[]
   *   An array of extracted URLs.
   */
  private function extractUrlsFromHtml($html): array {

    if (empty($html) || !is_string($html)) {
      return [];
    }

    $urls = [];

    // Load the HTML into a DOMDocument object for parsing.
    $dom = new \DOMDocument();
    libxml_use_internal_errors(TRUE);
    $dom->loadHTML($html, LIBXML_NOWARNING | LIBXML_NOERROR);
    libxml_clear_errors();

    // Get all anchor tags.
    foreach ($dom->getElementsByTagName('a') as $anchor) {
      if ($anchor->hasAttribute('href')) {
        $urls[] = $anchor->getAttribute('href');
      }
    }

    preg_match_all('/\b(?:https?:\/\/|www\.)[^\s"<]+/i', $html, $matches);
    if (!empty($matches[0])) {
      $urls = array_merge($urls, $matches[0]);
    }

    return array_unique($urls);
  }

}
